/* Enormous Fast Fourier Transform Library (EFFT) Version 1.0
   Copyright 2014, Colfax International

   Authors: ryo@colfax-intl.com     Ryo Asai
            andrey@colfax-intl.com  Andrey Vladimirov
            phi@colfax-intl.com     General inquiries

   For more information, refer to the publication
   Intel Cilk Plus for Complex Parallel Algorithms: "Enormous Fast Fourier Transforms" (EFFT) Library
   by Ryo Asai and Andrey Vladimirov (Colfax Research, 2014)
   available at http://research.colfaxinternational.com/post/2014/09/18/EFFT.aspx

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "efft.h"


//Reasembles an adjacent pair of Even and Odd bins
void ReassemblePair(float* array, const long size) {
  const float trigconst = -3.14159265359f/size; 

  // Special cases for 0-th and size-th element, since they do not have an imaginary part and is stored in a different way
  float* odds = &array[size];
  float* evens = &array[0];
  float even0 = evens[0];
  float odd0 = odds[0];
  float even1 = evens[1];
  float odd1 = odds[1];
  array[0]=even0+odd0;
  array[1]=even0-odd0;
  array[size]=even1;
  array[size+1]=-odd1;

  // Special case: size/2-th element, because it does not have a mirror element
  const long halfsize = size/2L; //Half the size of the EVEN and ODD arrays. Size is half the size of ARRAY 
  const float coshalf = 1.0f/sqrtf(2.0f);
  const float sinhalf = -coshalf;
  const float evenhalfre = evens[halfsize];
  const float evenhalfim = evens[halfsize+1];
  const float oddhalfsum = odds[halfsize]*sinhalf+odds[halfsize+1]*coshalf;
  const float oddhalfdif = odds[halfsize]*coshalf-odds[halfsize+1]*sinhalf;
  array[halfsize]=evenhalfre+oddhalfdif;
  array[halfsize+1]=evenhalfim+oddhalfsum;
  array[size+halfsize]=evenhalfre-oddhalfdif;
  array[size+halfsize+1]=-evenhalfim+oddhalfsum;

  //The first kTILE-1 iterations performed in a separate loop to avoid a conditional inside the main parallel loop
  //array[k] and array[2*size-k] are the k-th real and imaginary number given by the Cooley-Tukey method
  //array[size-k] and array[size+k] are real and imaginary numbers derived from the mirrored values of the k-th element.
  //The float values are temp holders for the array elements to allow for in-place recombination.
  const long kTILE = 64;
  
  for (long k = 1; k < kTILE; k++) {
    const float cosk = cosf(k*trigconst);
    const float sink = sinf(k*trigconst);

    const float evenre = evens[2*k  ];
    const float evenim = evens[2*k+1];
    const float oddsum = odds [2*k  ]*sink + odds[2*k+1]*cosk;
    const float odddif = odds [2*k  ]*cosk - odds[2*k+1]*sink;
  
    const float evenmirre =  evens[size-2*k  ];
    const float evenmirim =  evens[size-2*k+1];
    const float oddmirsum = -odds [size-2*k]*cosk - odds[size-2*k+1]*sink;
    const float oddmirdif = -odds [size-2*k]*sink + odds[size-2*k+1]*cosk;
  
    array[2*k]        =  evenre    + odddif;
    array[2*k+1]      =  evenim    + oddsum;
    array[size+2*k]   =  evenmirre - oddmirdif;
    array[size+2*k+1] = -evenmirim + oddmirsum;
  
    array[size-2*k]     =  evenmirre + oddmirdif;
    array[size-2*k+1]   =  evenmirim + oddmirsum;
    array[2*size-2*k]   =  evenre - odddif;
    array[2*size-2*k+1] = -evenim + oddsum;
  }

  //The main parallel loop. Does trigonometric function calcuation in a separate loop to enable automatic vectorization
#define ALIGNED __attribute__((aligned(64)))
  cilk_for (long kk = 2L*kTILE; kk < halfsize; kk+=2L*kTILE) {
    float 
      coslist[kTILE]    ALIGNED, sinlist[kTILE]    ALIGNED,
      evenrek[kTILE]    ALIGNED, evenimk[kTILE]    ALIGNED,
      evenmirrek[kTILE] ALIGNED, evenmirimk[kTILE] ALIGNED,
      oddrek[kTILE]     ALIGNED, oddimk[kTILE]     ALIGNED,
      oddmirrek[kTILE]  ALIGNED, oddmirimk[kTILE]  ALIGNED;

    const float theta = (float)(kk/2L)*trigconst; 

    // Vectorized twiddle factor precomputation
#pragma simd
#pragma vector aligned
    for (long i = 0; i < kTILE; i++) {
      sinlist[i] = sinf(theta+i*trigconst);
      coslist[i] = cosf(theta+i*trigconst);
    }

    // Gather stride-2 data for first pair
    evenrek[:] = evens[kk  :kTILE:2];
    evenimk[:] = evens[kk+1:kTILE:2];
    oddrek [:] = odds [kk  :kTILE:2];
    oddimk [:] = odds [kk+1:kTILE:2];

    evens[kk  :kTILE:2] = evenrek[:] + coslist[:]*oddrek[:] - sinlist[:]*oddimk[:];
    evens[kk+1:kTILE:2] = evenimk[:] + sinlist[:]*oddrek[:] + coslist[:]*oddimk[:];

    oddmirrek[:] = odds[size-kk  :kTILE:-2];
    oddmirimk[:] = odds[size-kk+1:kTILE:-2];

    odds[size-kk  :kTILE:-2] =  evenrek[:] - coslist[:]*oddrek[:] + sinlist[:]*oddimk[:];
    odds[size-kk+1:kTILE:-2] = -evenimk[:] + sinlist[:]*oddrek[:] + coslist[:]*oddimk[:];     

    evenmirrek[:] = evens[size-kk  :kTILE:-2];
    evenmirimk[:] = evens[size-kk+1:kTILE:-2];

    evens[size-kk  :kTILE:-2] =  evenmirrek[:] - sinlist[:]*oddmirrek[:] + coslist[:]*oddmirimk[:];
    evens[size-kk+1:kTILE:-2] =  evenmirimk[:] - coslist[:]*oddmirrek[:] - sinlist[:]*oddmirimk[:];
    odds [kk       :kTILE: 2] =  evenmirrek[:] + sinlist[:]*oddmirrek[:] - coslist[:]*oddmirimk[:];
    odds [kk+1     :kTILE: 2] = -evenmirimk[:] - coslist[:]*oddmirrek[:] - sinlist[:]*oddmirimk[:];

  }
}


void Scatter(float* dataArray, float* tempArray, const long binsize, const long numofbins, const long numofsplits) {
  long scatterIndex[numofbins];

  //bit_reversal permutation code. Determines which "bin" the element goes to
  for (long j = 0; j < 1L<< numofsplits; j++) {
    scatterIndex[j] = 0;
    for (long i = 0; i < numofsplits; i++)
      scatterIndex[j] += (j & 1L<<(numofsplits-i-1L)) ? 1L<<i : 0L;
  }

  const long iTILE=64L;
  cilk_for (long ii = 0; ii < binsize; ii += iTILE) {
    for(long j = 0; j < numofbins; j++) {
      for (long i = ii; i < ii+iTILE; i++) {
  	//Scatter_index is index mapping array, so that EVENS and ODDS DFT of the Cooley-Tukey method
  	// is next to each other. Algorithm-wise, this scatter does a bit-reverse permutation of the bottom {numofsplits}
  	// bits, shifts the remaining bits right by {numofsplits} and places the bit-reversed bottom bits at the top. 
  	tempArray[scatterIndex[j]*binsize+i] = dataArray[i*numofbins+j];
      }
    }
  }
}

void ProcessAndReassemble(DFTI_DESCRIPTOR_HANDLE* fftHandle, float* temp, const long n, const long binsize) {
  
  const long size = n/2;
  if (n > binsize) {

    // Recursively process half of array, parallelize recursion
    cilk_spawn ProcessAndReassemble(fftHandle, temp, size, binsize);
    ProcessAndReassemble(fftHandle, &temp[size], size, binsize);
    cilk_sync;

    // Apply the Cooley-Tukey reassembly to the pair of smaller processed FFTs
    ReassemblePair(temp,size);
    
  } else {
    // Perform direct DFT calculation using the single-threaded MKL DFT implementation
    long wid = __cilkrts_get_worker_number();
    DftiComputeForward(fftHandle[wid], temp);
  }
}



//////////////////////////////EFFT CLASS//////////////////////////////////////////////
EFFT_Transform::EFFT_Transform(const long N, const long splits) : size(N), numofsplits(splits) {
  
  //initiallizing the object-specific variables
  const long numofbins = 1L<<numofsplits;
  const long binsize = size/numofbins;
  const long numthreads = __cilkrts_get_nworkers();
  
  // Checking to make certain the bins will have the correct alignment. For 64-byte alignment with 4-byte float, each bin must
  // be a multiple of 64/4=16. To change this, go to the header file.
  const long excess = binsize % 256L;
  if(excess != 0) {
    printf("#ERROR n must be a multiple of %d ((2^{number of splits})*{256})\n", numofbins*256L);
    exit(1);
  }
  
  //Data allocation
  dataPtr = (float*) _mm_malloc(size*sizeof(float), 64);
  scratchPtr = (float*) _mm_malloc(size*sizeof(float), 64);

  DFTI_DESCRIPTOR_HANDLE* fftHandle = (DFTI_DESCRIPTOR_HANDLE*) malloc(numthreads*sizeof(DFTI_DESCRIPTOR_HANDLE));
  int checkInit[numthreads]; checkInit[:]=0;
  fftPlan = (void*) fftHandle;
 
  // Creating an MKL DFT handle for each worker
  cilk_for (long i = 0; i < numthreads; i++){
    long wid = __cilkrts_get_worker_number();
    mkl_set_num_threads(1);
    MKL_LONG fftsize = binsize;

    DftiCreateDescriptor(&fftHandle[wid], DFTI_SINGLE, DFTI_REAL, 1, fftsize);
    DftiSetValue(fftHandle[wid], DFTI_NUMBER_OF_USER_THREADS, 1);
    DftiSetValue(fftHandle[wid], DFTI_PACKED_FORMAT, DFTI_PERM_FORMAT);
    DftiCommitDescriptor (fftHandle[wid]);

    checkInit[wid] = 1;
  }
  
  // Checking that parallel initialization has succeeded
  bool ok = true;
  for (long i = 0; i < numthreads; i++)
    if (checkInit[i]!=1) ok=false;
  if (!ok) {
    printf("#ERROR Cilk Plus does not spawn as many workers as requested\n");
    exit(2);
  }
}

EFFT_Transform::~EFFT_Transform() {
  _mm_free(dataPtr);
  _mm_free(scratchPtr);
  free(fftPlan);
}

void EFFT_Transform::RunTransform() {

  const long numofbins = 1L<<numofsplits;

  //Scatter phase. Does a partial bit-reverse permute to scatter the full DFT into 2^{numofsplits} smaller DFTs ("bins")
  //This is done out-of-place, and, the data is put into the tempArray
  Scatter(dataPtr, scratchPtr, size/numofbins, numofbins, numofsplits);

  // Starting parallel recursion to carry out the small DFTs and reassemble them into the result
  ProcessAndReassemble((DFTI_DESCRIPTOR_HANDLE*)fftPlan, scratchPtr, size, size/numofbins);
}
