if [[ -n $1 && ( $1 == -h || $1 == --help ) ]]; then
    echo ""
    echo Runs a basic performance benchmark of the EFFT library. Note that
    echo '"benchmarkdriver"' is needed to run this script. Use '"make benchmark"' 
    echo to generate it.
    echo ""
    echo usage: runme.sh [option]
    echo ""
    echo options:
    echo ""
    echo "  -a  : "Automatically attempt to find good parameters for the
    echo "          "system. May not work for all systems.
    echo ""
    echo "  -nq : "Disables all queries. Neccessary inputs are taken as
    echo "          "arguments. Usage:
    echo "            "runme.sh -nq {Transform Sizes} {Workers List} "\\" 
    echo "                              "{Min Splits} {Max Splits} "\\"
    echo "                              "{Number of Trials}
    echo ""
    echo With no options, the script will query for the neccessary inputs.
    echo The inputs are:
    echo ""
    echo "   Transform Size-   "The number of elements to be Transformed.
    echo "                       "User may specify a list seperated by spaces
    echo "                       "to benchmark multiple sizes \(put quotes around the list\).
    echo ""
    echo "   Workers List-     "The number of Cilk Plus workers to be used.
    echo "                       "Generally, the best performance is gained when
    echo "                       "number of workers is equal to either the number
    echo "                       "of physical cores, or to a power of 2
    echo "                       "that is less than the number of cores.
    echo "                       "For example, in a 10-core system, try 8 and 10 workers.
    echo "                       "User may specify a list separated by spaces
    echo "                       "to benchmark different numbers of workers
    echo "                       "\(put quotes around the list\).
    echo ""
    echo "   Min Splits-       "Minimum number of splits to benchmark. Generally,
    echo "                       "the best performance is gained when "2^{splits}"
    echo "                       "is near the number of workers.
    echo ""
    echo "   Max Splits-       "Maximum number of splits to benchmark. Generally,
    echo "                       "the best performance is gained when "2^{splits}" 
    echo "                       "is near the number of workers.
    echo ""
    echo "   Number of Trials- "The number of elements to be Transformed.
    echo "                       "User may specify a list seperated by space
    echo "                       "to benchmark multiple sizes.
    echo ""
    echo "Example: runme.sh -nq \"33554432 268435456\" \"16 24\" 4 6 10"
    echo ""
    echo ""
    echo "For some systems, the script can automatically find the neccessary"
    echo "inputs. Run with -a option to mkake the script automatically find"
    echo "good values for the inputs."
    echo ""
    echo "Example: runme.sh -a"
    echo ""
    exit
fi

if [[ !(-e benchmarkdriver) ]]; then
    echo Missing benchmarkdriver. Use '"make benchmark"' to generate it.
    exit
fi


if [[ -n $1 &&  $1 == -a  ]]; then
if [[ !(-e /proc/cpuinfo) || !(-e /proc/meminfo) ]]; then
    echo Automatic parameter finding failed. 
    exit
fi
echo Attempting to automatically find good parameters to benchmark.
#GENERATING TEST WORKER NUMBERS
PHYS_CORES=`cat /proc/cpuinfo | egrep "core id|physical id" | awk 'BEGIN{V=0}{printf " %d", $4; if(V==1){printf "\n"}; V=1-V}' | sort | uniq | wc -l`

T=1
while [[ $(( 2 * $T )) -lt $PHYS_CORES ]]; do
    T=$(( $T * 2 ))
done

WORKERSLIST=$T" "$PHYS_CORES


#DETERMINING SPLITS
MAXTHREAD=$PHYS_CORES

S=1
while [[ $(( 2 ** $S )) -lt $(( $T / 2 )) ]]; do
    S=$(( $S + 1 ))
done
MINSPLITS=$S

while [[ $(( 2 ** $S )) -le $MAXTHREAD ]]; do
    S=$(( $S + 1 ))
done
MAXSPLITS=$S


#DETERMINING SIZE
MAXMEM=`cat /proc/meminfo | grep MemTotal: | awk '{print $2}'`
MAXMEM=$(( $MAXMEM * 1000 ))
#echo $MAXMEM
PWR=$MAXSPLITS
while [[ $(( 16 * ( 2 ** ( $PWR - $MINSPLITS ) ) * $MAXTHREAD )) -le $MAXMEM ]];do
    PWR=$(( $PWR + 1 ))
done
PWR=$(( $PWR - 1 ))
SIZELIST=$(( 2 ** $PWR ))
NTRIALS=10

echo Transform size: 2^$PWR = $SIZELIST
echo Using Workers: $WORKERSLIST
echo Trying split range: $MINSPLITS $MAXSPLITS
echo Trials per parameter set: $NTRIALS
echo ""
echo Hit enter to continue. q to quit.
read quit
if [[ $quit == "q" ]]; then exit; fi

elif [[ -n $1 && $1 == -nq ]]; then
    if [[ -z $2 || -z $3 || -z $4 || -z $5 || -z $6 ]]; then
	echo Missing arguments for -nq option. Format:
	echo runme.sh -nq {Transform Sizes} {Workers List} {Min Splits} {Max Splits} {Number of Trials}
	echo ""
	echo Example: runme.sh -nq "33554432 268435456" "16 24" 4 6 10
	exit
    fi
    SIZELIST=$2
    WORKERSLIST=$3
    MINSPLITS=$4
    MAXSPLITS=$5
    NTRIALS=$6

else
    echo ""
    echo Run \"$0 -a\" to get the script to attempt to automatically find good parameters for this system.
    echo Run \"$0 -h\" to get help on other parameters that this script accepts.
    echo ""
    echo No arguments were given, so we will perform Manual Parameter input.
    echo Hit enter to continue. q to quit.
    read quit
    if [[ $quit == "q" ]]; then exit; fi
    echo Please input Transform Size. Accepts multiple values separated by space. Example: 33554432 268435456 
    read SIZELIST
    echo Please input Workers List. Accepts multiple values separated by space. Example: 16 24 
    read WORKERSLIST
    echo Please input Min Splits. Example: 4
    read MINSPLITS
    echo Please input Max Splits. Example: 6
    read MAXSPLITS
    echo Please input Number of Trials. Example: 10 
    read NTRIALS
    echo Transform size\(s\): $SIZELIST
    echo Using Workers: $WORKERSLIST
    echo Trying split range: $MINSPLITS $MAXSPLITS
    echo Trials per parameter set: $NTRIALS
    echo ""
    echo Hit enter to continue. q to quit.
    read quit
    if [[ $quit == "q" ]]; then exit; fi
fi
for size in $SIZELIST; do 
    for splits in `seq $MINSPLITS $MAXSPLITS`; do
	for workers in $WORKERSLIST; do
	    export CILK_NWORKERS=$workers
	    export KMP_AFFINITY=scatter
            ./benchmarkdriver $size $splits $NTRIALS
	done
    done
done
